
.. DO NOT EDIT.
.. THIS FILE WAS AUTOMATICALLY GENERATED BY SPHINX-GALLERY.
.. TO MAKE CHANGES, EDIT THE SOURCE PYTHON FILE:
.. "generated/autoexamples/GPU/example_learn_samples.py"
.. LINE NUMBERS ARE GIVEN BELOW.

.. only:: html

    .. note::
        :class: sphx-glr-download-link-note

        :ref:`Go to the end <sphx_glr_download_generated_autoexamples_GPU_example_learn_samples.py>`
        to download the full example code. or to run this example in your browser via Binder

.. rst-class:: sphx-glr-example-title

.. _sphx_glr_generated_autoexamples_GPU_example_learn_samples.py:


======================
Learn Sampling pattern
======================

A small pytorch example to showcase learning k-space sampling patterns.
This example showcases the auto-diff capabilities of the NUFFT operator 
wrt to k-space trajectory in mri-nufft.

In this example, we solve the following optimization problem:

.. math::

    \mathbf{\hat{K}} =  \mathrm{arg} \min_{\mathbf{K}} ||  \mathcal{F}_\mathbf{K}^* D_\mathbf{K} \mathcal{F}_\mathbf{K} \mathbf{x} - \mathbf{x} ||_2^2
    
where :math:`\mathcal{F}_\mathbf{K}` is the forward NUFFT operator and :math:`D_\mathbf{K}` is the density compensators for trajectory :math:`\mathbf{K}`,  :math:`\mathbf{x}` is the MR image which is also the target image to be reconstructed.

.. warning::
    This example only showcases the autodiff capabilities, the learned sampling pattern is not scanner compliant as the scanner gradients required to implement it violate the hardware constraints. In practice, a projection :math:`\Pi_\mathcal{Q}(\mathbf{K})` into the scanner constraints set :math:`\mathcal{Q}` is recommended (see [Proj]_). This is implemented in the proprietary SPARKLING package [Sparks]_. Users are encouraged to contact the authors if they want to use it.

.. GENERATED FROM PYTHON SOURCE LINES 22-35

.. code-block:: Python

    import time
    import joblib

    import brainweb_dl as bwdl
    import matplotlib.pyplot as plt
    import numpy as np
    import torch
    from tqdm import tqdm
    from PIL import Image, ImageSequence

    from mrinufft import get_operator
    from mrinufft.trajectories import initialize_2D_radial








.. GENERATED FROM PYTHON SOURCE LINES 36-41

Setup a simple class to learn trajectory
----------------------------------------
.. note::
    While we are only learning the NUFFT operator, we still need the gradient ``wrt_data=True`` to be setup in ``get_operator`` to have all the gradients computed correctly.
    See [Projector]_ for more details.

.. GENERATED FROM PYTHON SOURCE LINES 41-70

.. code-block:: Python



    class Model(torch.nn.Module):
        def __init__(self, inital_trajectory):
            super(Model, self).__init__()
            self.trajectory = torch.nn.Parameter(
                data=torch.Tensor(inital_trajectory),
                requires_grad=True,
            )
            self.operator = get_operator("gpunufft", wrt_data=True, wrt_traj=True)(
                self.trajectory.detach().cpu().numpy(),
                shape=(256, 256),
                density=True,
                squeeze_dims=False,
            )

        def forward(self, x):
            # Update the trajectory in the NUFFT operator.
            # Note that the re-computation of density compensation happens internally.
            self.operator.samples = self.trajectory.clone()

            # A simple acquisition model simulated with a forward NUFFT operator
            kspace = self.operator.op(x)

            # A simple density compensated adjoint operator
            adjoint = self.operator.adj_op(kspace)
            return adjoint / torch.linalg.norm(adjoint)









.. GENERATED FROM PYTHON SOURCE LINES 71-73

Util function to plot the state of the model
--------------------------------------------

.. GENERATED FROM PYTHON SOURCE LINES 73-96

.. code-block:: Python



    def plot_state(axs, mri_2D, traj, recon, loss=None, save_name=None):
        axs = axs.flatten()
        axs[0].imshow(np.abs(mri_2D[0]), cmap="gray")
        axs[0].axis("off")
        axs[0].set_title("MR Image")
        axs[1].scatter(*traj.T, s=1)
        axs[1].set_title("Trajectory")
        axs[2].imshow(np.abs(recon[0][0].detach().cpu().numpy()), cmap="gray")
        axs[2].axis("off")
        axs[2].set_title("Reconstruction")
        if loss is not None:
            axs[3].plot(loss)
            axs[3].set_title("Loss")
            axs[3].grid("on")
        if save_name is not None:
            plt.savefig(save_name, bbox_inches="tight")
            plt.close()
        else:
            plt.show()









.. GENERATED FROM PYTHON SOURCE LINES 97-99

Setup model and optimizer
-------------------------

.. GENERATED FROM PYTHON SOURCE LINES 99-106

.. code-block:: Python

    init_traj = initialize_2D_radial(16, 512).reshape(-1, 2).astype(np.float32)
    model = Model(init_traj)
    optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)
    schedulder = torch.optim.lr_scheduler.LinearLR(
        optimizer, start_factor=1, end_factor=0.1, total_iters=100
    )








.. GENERATED FROM PYTHON SOURCE LINES 107-109

Setup data
----------

.. GENERATED FROM PYTHON SOURCE LINES 109-119

.. code-block:: Python


    mri_2D = torch.Tensor(np.flipud(bwdl.get_mri(4, "T1")[80, ...]).astype(np.complex64))[
        None
    ]
    mri_2D = mri_2D / torch.linalg.norm(mri_2D)
    model.eval()
    recon = model(mri_2D)
    fig, axs = plt.subplots(1, 3, figsize=(15, 5))
    plot_state(axs, mri_2D, init_traj, recon)




.. image-sg:: /generated/autoexamples/GPU/images/sphx_glr_example_learn_samples_001.png
   :alt: MR Image, Trajectory, Reconstruction
   :srcset: /generated/autoexamples/GPU/images/sphx_glr_example_learn_samples_001.png
   :class: sphx-glr-single-img





.. GENERATED FROM PYTHON SOURCE LINES 120-122

Start training loop
-------------------

.. GENERATED FROM PYTHON SOURCE LINES 122-168

.. code-block:: Python

    losses = []
    image_files = []
    model.train()
    with tqdm(range(100), unit="steps") as tqdms:
        for i in tqdms:
            out = model(mri_2D)
            loss = torch.norm(out - mri_2D[None])
            numpy_loss = loss.detach().cpu().numpy()
            tqdms.set_postfix({"loss": numpy_loss})
            losses.append(numpy_loss)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            with torch.no_grad():
                # Clamp the value of trajectory between [-0.5, 0.5]
                for param in model.parameters():
                    param.clamp_(-0.5, 0.5)
            schedulder.step()
            # Generate images for gif
            hashed = joblib.hash((i, "learn_traj", time.time()))
            filename = "/tmp/" + f"{hashed}.png"
            fig, axs = plt.subplots(2, 2, figsize=(10, 10))
            plot_state(
                axs,
                mri_2D,
                model.trajectory.detach().cpu().numpy(),
                out,
                losses,
                save_name=filename,
            )
            image_files.append(filename)


    # Make a GIF of all images.
    imgs = [Image.open(img) for img in image_files]
    imgs[0].save(
        "mrinufft_learn_traj.gif",
        save_all=True,
        append_images=imgs[1:],
        optimize=False,
        duration=2,
        loop=0,
    )

    # sphinx_gallery_thumbnail_path = 'generated/autoexamples/GPU/images/mrinufft_learn_traj.gif'





.. rst-class:: sphx-glr-script-out

 .. code-block:: none

      0%|          | 0/100 [00:00<?, ?steps/s]      0%|          | 0/100 [00:00<?, ?steps/s, loss=0.29853177]      1%|          | 1/100 [00:00<00:58,  1.70steps/s, loss=0.29853177]      1%|          | 1/100 [00:00<00:58,  1.70steps/s, loss=0.37153634]      2%|▏         | 2/100 [00:01<00:56,  1.72steps/s, loss=0.37153634]      2%|▏         | 2/100 [00:01<00:56,  1.72steps/s, loss=0.3289368]       3%|▎         | 3/100 [00:01<00:57,  1.70steps/s, loss=0.3289368]      3%|▎         | 3/100 [00:01<00:57,  1.70steps/s, loss=0.27461457]      4%|▍         | 4/100 [00:02<00:55,  1.73steps/s, loss=0.27461457]      4%|▍         | 4/100 [00:02<00:55,  1.73steps/s, loss=0.2835944]       5%|▌         | 5/100 [00:02<00:53,  1.76steps/s, loss=0.2835944]      5%|▌         | 5/100 [00:02<00:53,  1.76steps/s, loss=0.25903752]      6%|▌         | 6/100 [00:03<00:53,  1.74steps/s, loss=0.25903752]      6%|▌         | 6/100 [00:03<00:53,  1.74steps/s, loss=0.25020278]      7%|▋         | 7/100 [00:04<00:52,  1.76steps/s, loss=0.25020278]      7%|▋         | 7/100 [00:04<00:52,  1.76steps/s, loss=0.24855691]      8%|▊         | 8/100 [00:04<00:57,  1.59steps/s, loss=0.24855691]      8%|▊         | 8/100 [00:04<00:57,  1.59steps/s, loss=0.23702781]      9%|▉         | 9/100 [00:05<00:55,  1.65steps/s, loss=0.23702781]      9%|▉         | 9/100 [00:05<00:55,  1.65steps/s, loss=0.23232004]     10%|█         | 10/100 [00:05<00:53,  1.69steps/s, loss=0.23232004]     10%|█         | 10/100 [00:06<00:53,  1.69steps/s, loss=0.22273493]     11%|█         | 11/100 [00:06<00:54,  1.63steps/s, loss=0.22273493]     11%|█         | 11/100 [00:06<00:54,  1.63steps/s, loss=0.220687]       12%|█▏        | 12/100 [00:07<00:52,  1.67steps/s, loss=0.220687]     12%|█▏        | 12/100 [00:07<00:52,  1.67steps/s, loss=0.21408738]     13%|█▎        | 13/100 [00:07<00:51,  1.69steps/s, loss=0.21408738]     13%|█▎        | 13/100 [00:07<00:51,  1.69steps/s, loss=0.20796785]     14%|█▍        | 14/100 [00:08<00:50,  1.69steps/s, loss=0.20796785]     14%|█▍        | 14/100 [00:08<00:50,  1.69steps/s, loss=0.20363936]     15%|█▌        | 15/100 [00:08<00:49,  1.70steps/s, loss=0.20363936]     15%|█▌        | 15/100 [00:08<00:49,  1.70steps/s, loss=0.19725086]     16%|█▌        | 16/100 [00:09<00:49,  1.70steps/s, loss=0.19725086]     16%|█▌        | 16/100 [00:09<00:49,  1.70steps/s, loss=0.19570541]     17%|█▋        | 17/100 [00:10<00:54,  1.52steps/s, loss=0.19570541]     17%|█▋        | 17/100 [00:10<00:54,  1.52steps/s, loss=0.19244406]     18%|█▊        | 18/100 [00:10<00:52,  1.57steps/s, loss=0.19244406]     18%|█▊        | 18/100 [00:10<00:52,  1.57steps/s, loss=0.1893066]      19%|█▉        | 19/100 [00:11<00:50,  1.61steps/s, loss=0.1893066]     19%|█▉        | 19/100 [00:11<00:50,  1.61steps/s, loss=0.18791121]     20%|██        | 20/100 [00:12<00:48,  1.65steps/s, loss=0.18791121]     20%|██        | 20/100 [00:12<00:48,  1.65steps/s, loss=0.18381003]     21%|██        | 21/100 [00:12<00:47,  1.66steps/s, loss=0.18381003]     21%|██        | 21/100 [00:12<00:47,  1.66steps/s, loss=0.18035991]     22%|██▏       | 22/100 [00:13<00:46,  1.67steps/s, loss=0.18035991]     22%|██▏       | 22/100 [00:13<00:46,  1.67steps/s, loss=0.1803372]      23%|██▎       | 23/100 [00:13<00:45,  1.68steps/s, loss=0.1803372]     23%|██▎       | 23/100 [00:13<00:45,  1.68steps/s, loss=0.17593923]     24%|██▍       | 24/100 [00:14<00:44,  1.70steps/s, loss=0.17593923]     24%|██▍       | 24/100 [00:14<00:44,  1.70steps/s, loss=0.17714013]     25%|██▌       | 25/100 [00:15<00:48,  1.54steps/s, loss=0.17714013]     25%|██▌       | 25/100 [00:15<00:48,  1.54steps/s, loss=0.17497817]     26%|██▌       | 26/100 [00:15<00:47,  1.57steps/s, loss=0.17497817]     26%|██▌       | 26/100 [00:15<00:47,  1.57steps/s, loss=0.17569448]     27%|██▋       | 27/100 [00:16<00:45,  1.61steps/s, loss=0.17569448]     27%|██▋       | 27/100 [00:16<00:45,  1.61steps/s, loss=0.17163812]     28%|██▊       | 28/100 [00:16<00:43,  1.64steps/s, loss=0.17163812]     28%|██▊       | 28/100 [00:16<00:43,  1.64steps/s, loss=0.17519206]     29%|██▉       | 29/100 [00:17<00:42,  1.66steps/s, loss=0.17519206]     29%|██▉       | 29/100 [00:17<00:42,  1.66steps/s, loss=0.17623131]     30%|███       | 30/100 [00:18<00:41,  1.68steps/s, loss=0.17623131]     30%|███       | 30/100 [00:18<00:41,  1.68steps/s, loss=0.17288709]     31%|███       | 31/100 [00:18<00:41,  1.68steps/s, loss=0.17288709]     31%|███       | 31/100 [00:18<00:41,  1.68steps/s, loss=0.17123584]     32%|███▏      | 32/100 [00:19<00:40,  1.68steps/s, loss=0.17123584]     32%|███▏      | 32/100 [00:19<00:40,  1.68steps/s, loss=0.17045005]     33%|███▎      | 33/100 [00:19<00:39,  1.68steps/s, loss=0.17045005]     33%|███▎      | 33/100 [00:19<00:39,  1.68steps/s, loss=0.170718]       34%|███▍      | 34/100 [00:20<00:43,  1.52steps/s, loss=0.170718]     34%|███▍      | 34/100 [00:20<00:43,  1.52steps/s, loss=0.17444049]     35%|███▌      | 35/100 [00:21<00:41,  1.57steps/s, loss=0.17444049]     35%|███▌      | 35/100 [00:21<00:41,  1.57steps/s, loss=0.17083822]     36%|███▌      | 36/100 [00:21<00:41,  1.56steps/s, loss=0.17083822]     36%|███▌      | 36/100 [00:21<00:41,  1.56steps/s, loss=0.16629073]     37%|███▋      | 37/100 [00:22<00:39,  1.61steps/s, loss=0.16629073]     37%|███▋      | 37/100 [00:22<00:39,  1.61steps/s, loss=0.17087366]     38%|███▊      | 38/100 [00:23<00:37,  1.67steps/s, loss=0.17087366]     38%|███▊      | 38/100 [00:23<00:37,  1.67steps/s, loss=0.1676508]      39%|███▉      | 39/100 [00:23<00:35,  1.72steps/s, loss=0.1676508]     39%|███▉      | 39/100 [00:23<00:35,  1.72steps/s, loss=0.17123078]     40%|████      | 40/100 [00:24<00:34,  1.75steps/s, loss=0.17123078]     40%|████      | 40/100 [00:24<00:34,  1.75steps/s, loss=0.1695712]      41%|████      | 41/100 [00:24<00:33,  1.77steps/s, loss=0.1695712]     41%|████      | 41/100 [00:24<00:33,  1.77steps/s, loss=0.16771725]     42%|████▏     | 42/100 [00:25<00:36,  1.60steps/s, loss=0.16771725]     42%|████▏     | 42/100 [00:25<00:36,  1.60steps/s, loss=0.17046179]     43%|████▎     | 43/100 [00:25<00:34,  1.66steps/s, loss=0.17046179]     43%|████▎     | 43/100 [00:26<00:34,  1.66steps/s, loss=0.17307211]     44%|████▍     | 44/100 [00:26<00:32,  1.72steps/s, loss=0.17307211]     44%|████▍     | 44/100 [00:26<00:32,  1.72steps/s, loss=0.1677222]      45%|████▌     | 45/100 [00:27<00:31,  1.75steps/s, loss=0.1677222]     45%|████▌     | 45/100 [00:27<00:31,  1.75steps/s, loss=0.17208856]     46%|████▌     | 46/100 [00:27<00:30,  1.78steps/s, loss=0.17208856]     46%|████▌     | 46/100 [00:27<00:30,  1.78steps/s, loss=0.17628533]     47%|████▋     | 47/100 [00:28<00:30,  1.75steps/s, loss=0.17628533]     47%|████▋     | 47/100 [00:28<00:30,  1.75steps/s, loss=0.1693144]      48%|████▊     | 48/100 [00:28<00:29,  1.77steps/s, loss=0.1693144]     48%|████▊     | 48/100 [00:28<00:29,  1.77steps/s, loss=0.16934024]     49%|████▉     | 49/100 [00:29<00:28,  1.79steps/s, loss=0.16934024]     49%|████▉     | 49/100 [00:29<00:28,  1.79steps/s, loss=0.1802284]      50%|█████     | 50/100 [00:29<00:27,  1.80steps/s, loss=0.1802284]     50%|█████     | 50/100 [00:29<00:27,  1.80steps/s, loss=0.17053309]     51%|█████     | 51/100 [00:30<00:30,  1.62steps/s, loss=0.17053309]     51%|█████     | 51/100 [00:30<00:30,  1.62steps/s, loss=0.16841577]     52%|█████▏    | 52/100 [00:31<00:29,  1.62steps/s, loss=0.16841577]     52%|█████▏    | 52/100 [00:31<00:29,  1.62steps/s, loss=0.18361828]     53%|█████▎    | 53/100 [00:31<00:28,  1.67steps/s, loss=0.18361828]     53%|█████▎    | 53/100 [00:31<00:28,  1.67steps/s, loss=0.1729434]      54%|█████▍    | 54/100 [00:32<00:26,  1.71steps/s, loss=0.1729434]     54%|█████▍    | 54/100 [00:32<00:26,  1.71steps/s, loss=0.17125648]     55%|█████▌    | 55/100 [00:32<00:25,  1.75steps/s, loss=0.17125648]     55%|█████▌    | 55/100 [00:32<00:25,  1.75steps/s, loss=0.17031834]     56%|█████▌    | 56/100 [00:33<00:25,  1.76steps/s, loss=0.17031834]     56%|█████▌    | 56/100 [00:33<00:25,  1.76steps/s, loss=0.15803307]     57%|█████▋    | 57/100 [00:34<00:24,  1.76steps/s, loss=0.15803307]     57%|█████▋    | 57/100 [00:34<00:24,  1.76steps/s, loss=0.16030627]     58%|█████▊    | 58/100 [00:34<00:23,  1.77steps/s, loss=0.16030627]     58%|█████▊    | 58/100 [00:34<00:23,  1.77steps/s, loss=0.1601321]      59%|█████▉    | 59/100 [00:35<00:23,  1.77steps/s, loss=0.1601321]     59%|█████▉    | 59/100 [00:35<00:23,  1.77steps/s, loss=0.15379326]     60%|██████    | 60/100 [00:35<00:24,  1.60steps/s, loss=0.15379326]     60%|██████    | 60/100 [00:35<00:24,  1.60steps/s, loss=0.16112208]     61%|██████    | 61/100 [00:36<00:23,  1.64steps/s, loss=0.16112208]     61%|██████    | 61/100 [00:36<00:23,  1.64steps/s, loss=0.15892589]     62%|██████▏   | 62/100 [00:37<00:22,  1.67steps/s, loss=0.15892589]     62%|██████▏   | 62/100 [00:37<00:22,  1.67steps/s, loss=0.15204129]     63%|██████▎   | 63/100 [00:37<00:21,  1.69steps/s, loss=0.15204129]     63%|██████▎   | 63/100 [00:37<00:21,  1.69steps/s, loss=0.15557532]     64%|██████▍   | 64/100 [00:38<00:21,  1.71steps/s, loss=0.15557532]     64%|██████▍   | 64/100 [00:38<00:21,  1.71steps/s, loss=0.14926282]     65%|██████▌   | 65/100 [00:38<00:20,  1.73steps/s, loss=0.14926282]     65%|██████▌   | 65/100 [00:38<00:20,  1.73steps/s, loss=0.14790568]     66%|██████▌   | 66/100 [00:39<00:19,  1.73steps/s, loss=0.14790568]     66%|██████▌   | 66/100 [00:39<00:19,  1.73steps/s, loss=0.15179561]     67%|██████▋   | 67/100 [00:39<00:18,  1.75steps/s, loss=0.15179561]     67%|██████▋   | 67/100 [00:39<00:18,  1.75steps/s, loss=0.15050074]     68%|██████▊   | 68/100 [00:40<00:18,  1.74steps/s, loss=0.15050074]     68%|██████▊   | 68/100 [00:40<00:18,  1.74steps/s, loss=0.1476706]      69%|██████▉   | 69/100 [00:41<00:19,  1.55steps/s, loss=0.1476706]     69%|██████▉   | 69/100 [00:41<00:19,  1.55steps/s, loss=0.14766733]     70%|███████   | 70/100 [00:41<00:18,  1.61steps/s, loss=0.14766733]     70%|███████   | 70/100 [00:41<00:18,  1.61steps/s, loss=0.14772314]     71%|███████   | 71/100 [00:42<00:17,  1.64steps/s, loss=0.14772314]     71%|███████   | 71/100 [00:42<00:17,  1.64steps/s, loss=0.14791068]     72%|███████▏  | 72/100 [00:42<00:16,  1.68steps/s, loss=0.14791068]     72%|███████▏  | 72/100 [00:43<00:16,  1.68steps/s, loss=0.14307731]     73%|███████▎  | 73/100 [00:43<00:15,  1.70steps/s, loss=0.14307731]     73%|███████▎  | 73/100 [00:43<00:15,  1.70steps/s, loss=0.14209394]     74%|███████▍  | 74/100 [00:44<00:15,  1.72steps/s, loss=0.14209394]     74%|███████▍  | 74/100 [00:44<00:15,  1.72steps/s, loss=0.14288136]     75%|███████▌  | 75/100 [00:44<00:14,  1.73steps/s, loss=0.14288136]     75%|███████▌  | 75/100 [00:44<00:14,  1.73steps/s, loss=0.14284274]     76%|███████▌  | 76/100 [00:45<00:13,  1.77steps/s, loss=0.14284274]     76%|███████▌  | 76/100 [00:45<00:13,  1.77steps/s, loss=0.14195457]     77%|███████▋  | 77/100 [00:45<00:12,  1.78steps/s, loss=0.14195457]     77%|███████▋  | 77/100 [00:45<00:12,  1.78steps/s, loss=0.14378512]     78%|███████▊  | 78/100 [00:46<00:13,  1.61steps/s, loss=0.14378512]     78%|███████▊  | 78/100 [00:46<00:13,  1.61steps/s, loss=0.14391056]     79%|███████▉  | 79/100 [00:47<00:12,  1.64steps/s, loss=0.14391056]     79%|███████▉  | 79/100 [00:47<00:12,  1.64steps/s, loss=0.1457778]      80%|████████  | 80/100 [00:47<00:11,  1.69steps/s, loss=0.1457778]     80%|████████  | 80/100 [00:47<00:11,  1.69steps/s, loss=0.14724785]     81%|████████  | 81/100 [00:48<00:11,  1.71steps/s, loss=0.14724785]     81%|████████  | 81/100 [00:48<00:11,  1.71steps/s, loss=0.15241282]     82%|████████▏ | 82/100 [00:48<00:10,  1.72steps/s, loss=0.15241282]     82%|████████▏ | 82/100 [00:48<00:10,  1.72steps/s, loss=0.14621902]     83%|████████▎ | 83/100 [00:49<00:09,  1.75steps/s, loss=0.14621902]     83%|████████▎ | 83/100 [00:49<00:09,  1.75steps/s, loss=0.14746276]     84%|████████▍ | 84/100 [00:49<00:09,  1.76steps/s, loss=0.14746276]     84%|████████▍ | 84/100 [00:49<00:09,  1.76steps/s, loss=0.14773417]     85%|████████▌ | 85/100 [00:50<00:08,  1.77steps/s, loss=0.14773417]     85%|████████▌ | 85/100 [00:50<00:08,  1.77steps/s, loss=0.14672548]     86%|████████▌ | 86/100 [00:51<00:07,  1.78steps/s, loss=0.14672548]     86%|████████▌ | 86/100 [00:51<00:07,  1.78steps/s, loss=0.14951788]     87%|████████▋ | 87/100 [00:51<00:08,  1.59steps/s, loss=0.14951788]     87%|████████▋ | 87/100 [00:51<00:08,  1.59steps/s, loss=0.15210204]     88%|████████▊ | 88/100 [00:52<00:07,  1.65steps/s, loss=0.15210204]     88%|████████▊ | 88/100 [00:52<00:07,  1.65steps/s, loss=0.1490561]      89%|████████▉ | 89/100 [00:52<00:06,  1.69steps/s, loss=0.1490561]     89%|████████▉ | 89/100 [00:52<00:06,  1.69steps/s, loss=0.14859655]     90%|█████████ | 90/100 [00:53<00:05,  1.71steps/s, loss=0.14859655]     90%|█████████ | 90/100 [00:53<00:05,  1.71steps/s, loss=0.14393204]     91%|█████████ | 91/100 [00:54<00:05,  1.74steps/s, loss=0.14393204]     91%|█████████ | 91/100 [00:54<00:05,  1.74steps/s, loss=0.14268024]     92%|█████████▏| 92/100 [00:54<00:04,  1.76steps/s, loss=0.14268024]     92%|█████████▏| 92/100 [00:54<00:04,  1.76steps/s, loss=0.14316559]     93%|█████████▎| 93/100 [00:55<00:03,  1.78steps/s, loss=0.14316559]     93%|█████████▎| 93/100 [00:55<00:03,  1.78steps/s, loss=0.14432459]     94%|█████████▍| 94/100 [00:55<00:03,  1.78steps/s, loss=0.14432459]     94%|█████████▍| 94/100 [00:55<00:03,  1.78steps/s, loss=0.14435548]     95%|█████████▌| 95/100 [00:56<00:02,  1.78steps/s, loss=0.14435548]     95%|█████████▌| 95/100 [00:56<00:02,  1.78steps/s, loss=0.15041222]     96%|█████████▌| 96/100 [00:57<00:02,  1.61steps/s, loss=0.15041222]     96%|█████████▌| 96/100 [00:57<00:02,  1.61steps/s, loss=0.1505817]      97%|█████████▋| 97/100 [00:57<00:01,  1.66steps/s, loss=0.1505817]     97%|█████████▋| 97/100 [00:57<00:01,  1.66steps/s, loss=0.15185781]     98%|█████████▊| 98/100 [00:58<00:01,  1.69steps/s, loss=0.15185781]     98%|█████████▊| 98/100 [00:58<00:01,  1.69steps/s, loss=0.15383466]     99%|█████████▉| 99/100 [00:58<00:00,  1.72steps/s, loss=0.15383466]     99%|█████████▉| 99/100 [00:58<00:00,  1.72steps/s, loss=0.15410317]    100%|██████████| 100/100 [00:59<00:00,  1.71steps/s, loss=0.15410317]    100%|██████████| 100/100 [00:59<00:00,  1.69steps/s, loss=0.15410317]




.. GENERATED FROM PYTHON SOURCE LINES 195-199

.. image-sg:: /generated/autoexamples/GPU/images/mrinufft_learn_traj.gif
   :alt: example learn_samples
   :srcset: /generated/autoexamples/GPU/images/mrinufft_learn_traj.gif
   :class: sphx-glr-single-img

.. GENERATED FROM PYTHON SOURCE LINES 201-203

Trained trajectory
------------------

.. GENERATED FROM PYTHON SOURCE LINES 203-209

.. code-block:: Python

    model.eval()
    recon = model(mri_2D)
    fig, axs = plt.subplots(2, 2, figsize=(10, 10))
    plot_state(axs, mri_2D, model.trajectory.detach().cpu().numpy(), recon, losses)
    plt.show()




.. image-sg:: /generated/autoexamples/GPU/images/sphx_glr_example_learn_samples_002.png
   :alt: MR Image, Trajectory, Reconstruction, Loss
   :srcset: /generated/autoexamples/GPU/images/sphx_glr_example_learn_samples_002.png
   :class: sphx-glr-single-img





.. GENERATED FROM PYTHON SOURCE LINES 210-225

References
==========

.. [Proj] N. Chauffert, P. Weiss, J. Kahn and P. Ciuciu, "A Projection Algorithm for
          Gradient Waveforms Design in Magnetic Resonance Imaging," in
          IEEE Transactions on Medical Imaging, vol. 35, no. 9, pp. 2026-2039, Sept. 2016,
          doi: 10.1109/TMI.2016.2544251.
.. [Sparks] Chaithya GR, P. Weiss, G. Daval-Frérot, A. Massire, A. Vignaud and P. Ciuciu,
          "Optimizing Full 3D SPARKLING Trajectories for High-Resolution Magnetic
          Resonance Imaging," in IEEE Transactions on Medical Imaging, vol. 41, no. 8,
          pp. 2105-2117, Aug. 2022, doi: 10.1109/TMI.2022.3157269.
.. [Projector] Chaithya GR, and Philippe Ciuciu. 2023. "Jointly Learning Non-Cartesian
          k-Space Trajectories and Reconstruction Networks for 2D and 3D MR Imaging
          through Projection" Bioengineering 10, no. 2: 158.
          https://doi.org/10.3390/bioengineering10020158


.. rst-class:: sphx-glr-timing

   **Total running time of the script:** (1 minutes 4.214 seconds)


.. _sphx_glr_download_generated_autoexamples_GPU_example_learn_samples.py:

.. only:: html

  .. container:: sphx-glr-footer sphx-glr-footer-example

    .. container:: binder-badge

      .. image:: images/binder_badge_logo.svg
        :target: https://mybinder.org/v2/gh/mind-inria/mri-nufft/gh-pages?urlpath=lab/tree/examples/generated/autoexamples/GPU/example_learn_samples.ipynb
        :alt: Launch binder
        :width: 150 px

    .. container:: sphx-glr-download sphx-glr-download-jupyter

      :download:`Download Jupyter notebook: example_learn_samples.ipynb <example_learn_samples.ipynb>`

    .. container:: sphx-glr-download sphx-glr-download-python

      :download:`Download Python source code: example_learn_samples.py <example_learn_samples.py>`

    .. container:: sphx-glr-download sphx-glr-download-zip

      :download:`Download zipped: example_learn_samples.zip <example_learn_samples.zip>`


.. only:: html

 .. rst-class:: sphx-glr-signature

    `Gallery generated by Sphinx-Gallery <https://sphinx-gallery.github.io>`_
